/* This file is part of the Imote2 Codeloader.
 *
 * The Imote2 Codeloader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Imote2 Codeloader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Imote2 Codeloader. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * Packet.java
 *
 * @author Markus Engel
 * Copyright (c) 2009,2010 Networked Systems Group, University of Kaiserslautern
 */
package imote.communication;

import ch.ntb.usb.*;
import imote.communication.tx.Command;

/** Packet */
public class Packet {
	/** header size, must match bootloader */
	public static final int HEADER_SIZE = 4;
	/** data */
	private final byte[] rawData;
	/** data position */
	private int pos;
	/** data length */
	private int dataLength;
	/** message type identifier */
	private byte command;

	/** setup new packet for sending
	 *
	 * @param cmd message type identifier
	 * @param dataLength message payload length
	 * @param compressed true iff payload is compressed
	 */
	protected Packet(Command cmd, int dataLength, boolean compressed) {
		this.rawData = new byte[dataLength + HEADER_SIZE];
		this.pos = 0;
		this.dataLength = dataLength;
		this.command = cmd.getCmd();

		/* setup header */
		this.encodeUShort(this.dataLength + HEADER_SIZE);
		this.encodeUByte(this.command);
		if (compressed) {
			this.encodeUByte((byte)(1 << 0));
		} else {
			this.encodeUByte((byte)0);
		}
	}

	/** setup new packet for reception
	 * 
	 * @param dataLength payload length
	 */
	protected Packet(int dataLength) {
		this.rawData = new byte[dataLength + HEADER_SIZE];
	}

	/** get payload length
	 *
	 * @return payload length
	 */
	public int getDataLength() {
		return this.dataLength;
	}

	/** get message type identifier
	 *
	 * @return message type identifier
	 */
	public int getCmd() {
		return this.command;
	}

	/** send packet over USB
	 *
	 * @param device device to send packet to
	 * @param endpoint endpoint of device to send packet to
	 * @param timeout max timeout in ms
	 * @return number of bytes sent
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public int send(Device device, int endpoint, int timeout)
		throws USBException {
		int r = device.writeBulk(endpoint, this.rawData, this.rawData.length,
			timeout, false);
		if (this.rawData.length % 64 == 0) {
			/* send zlp */
			device.writeBulk(endpoint,null, 0, timeout, false);
		}
		return r;
	}

	/** receive packet from USB
	 *
	 * @param device device to receive from
	 * @param endpoint endpoint of device to receive from
	 * @param timeout timeout in ms
	 * @return number of bytes received
	 * @throws USBException whenever something goes wrong concerning USB
	 */
	public int recv(Device device, int endpoint, int timeout)
		throws USBException {
		int ret;

		ret = device.readBulk(endpoint, this.rawData, this.rawData.length,
			timeout, false);
		this.pos = 0;
		this.dataLength = this.decodeUShort() - HEADER_SIZE;
		this.command = (byte)this.decodeUByte();

		return ret;
	}

	/** append a byte array to payload
	 *
	 * @param data input data
	 * @param length length of data
	 */
	protected final void encodeByteArray(byte[] data, int length) {
		System.arraycopy(data, 0, this.rawData, this.pos, length);
		this.pos += length;
	}

	/** append a byte array to payload
	 *
	 * @param data input data
	 * @param offset offset in input data
	 * @param length length of data
	 */
	protected final void encodeByteArray(byte[] data, int offset, int length) {
		System.arraycopy(data, offset, this.rawData, this.pos, length);
		this.pos += length;
	}

	/** decode a byte array from the payload
	 *
	 * @param length length of data
	 * @return byte array
	 */
	protected final byte[] decodeByteArray(int length) {
		byte[] data = new byte[length];
		System.arraycopy(this.rawData, this.pos, data, 0, length);
		this.pos += length;
		return data;
	}

	/** append an unsigned byte to payload
	 *
	 * @param integer unsigned byte
	 */
	protected final void encodeUByte(short integer) {
		this.rawData[this.pos++] = (byte) ((integer >>> 0) & 0xff);
	}

	/** decode an unsigned byte from payload
	 *
	 * @return unsigned byte
	 */
	protected final short decodeUByte() {
		return (short) ((short) this.rawData[this.pos++] & 0xff);
	}

	/** append an unsigned short to payload
	 *
	 * @param integer unsigned short
	 */
	protected final void encodeUShort(int integer) {
		this.rawData[this.pos++] = (byte) ((integer >>> 0) & 0xff);
		this.rawData[this.pos++] = (byte) ((integer >>> 8) & 0xff);
	}

	/** decode an unsigned short from payload
	 *
	 * @return unsigned short
	 */
	protected final int decodeUShort() {
		return (((int) this.rawData[this.pos++] & 0xff) << 0)
			| (((int) this.rawData[this.pos++] & 0xff) << 8);
	}

	/** append an unsigned integer to payload
	 *
	 * @param integer unsigned integer
	 */
	protected final void encodeUInteger(long integer) {
		this.rawData[this.pos++] = (byte) ((integer >>> 0) & 0xff);
		this.rawData[this.pos++] = (byte) ((integer >>> 8) & 0xff);
		this.rawData[this.pos++] = (byte) ((integer >>> 16) & 0xff);
		this.rawData[this.pos++] = (byte) ((integer >>> 24) & 0xff);
	}

	/** decode an unsigned integer from the payload
	 *
	 * @return unsigned integer
	 */
	protected final long decodeUInteger() {
		return (((int) this.rawData[this.pos++] & 0xff) << 0)
			| (((int) this.rawData[this.pos++] & 0xff) << 8)
			| (((int) this.rawData[this.pos++] & 0xff) << 16)
			| (((int) this.rawData[this.pos++] & 0xff) << 24);
	}
	/*
	protected final void encodeInteger(int offset, int integer) {
	this.rawData[offset + 0] = (byte)((integer >>> 0) & 0xff);
	this.rawData[offset + 1] = (byte)((integer >>> 8) & 0xff);
	this.rawData[offset + 2] = (byte)((integer >>> 16) & 0xff);
	this.rawData[offset + 3] = (byte)((integer >>> 24) & 0xff);
	}

	protected final int decodeInteger(int offset) {
	return (this.rawData[offset + 0] << 0) |
	(this.rawData[offset + 1] << 8) |
	(this.rawData[offset + 2] << 16) |
	(this.rawData[offset + 3] << 24);
	}*/
}

